/* $Id$ */

/** @file headway_gui.cpp Headway control window. */

#include "stdafx.h"
#include "openttd.h"
#include "debug.h"
#include "company_func.h"
#include "gui.h"
#include "window_gui.h"
#include "textbuf_gui.h"
#include "command_func.h"
#include "command_type.h"
#include "variables.h"
#include "vehicle_gui.h"
#include "vehicle_gui_base.h"
#include "viewport_func.h"
#include "gfx_func.h"
#include "train.h"
#include "ship.h"
#include "aircraft.h"
#include "roadveh.h"
#include "depot_base.h"
#include "cargotype.h"
#include "group.h"
#include "group_gui.h"
#include "strings_func.h"
#include "functions.h"
#include "timetable_type.h"
#include "window_func.h"
#include "vehicle_func.h"
#include "autoreplace_gui.h"
#include "core/alloc_func.hpp"
#include "core/mem_func.hpp"
#include "core/sort_func.hpp"
#include "string_func.h"
#include "settings_type.h"
#include "widgets/dropdown_func.h"
#include "order_func.h"
#include "timetable.h"
#include "vehiclelist.h"
#include "date_func.h"
#include "date_gui.h"

#include "table/sprites.h"
#include "table/strings.h"


enum HeadwayListWindowWidgets {
	HLW_WIDGET_CLOSEBOX = 0,
	HLW_WIDGET_CAPTION,
	HLW_WIDGET_STICKY,
	HLW_WIDGET_EMPTY_TOP_LEFT,
	HLW_WIDGET_ORDER_LIST,
	HLW_WIDGET_FAKED_SCROLLBAR1, // Hidden, allows scrolling of panel
	HLW_WIDGET_VEH_LIST,
	HLW_WIDGET_FAKED_SCROLLBAR2, // Hidden, allows scrolling of panel
	HLW_WIDGET_GAP_LIST,
	HLW_WIDGET_SCROLLBAR,
	HLW_WIDGET_EMPTY_GAP,
	HLW_WIDGET_INFO_PANEL,

	HLW_WIDGET_INTERVAL,
	HLW_WIDGET_AUTOMATIC,
	HLW_WIDGET_RESET_LATENESS,

	HLW_WIDGET_STOP_ALL,
	HLW_WIDGET_START_ALL,
	HLW_WIDGET_EMPTY_BOTTOM_RIGHT,
	HLW_WIDGET_RESIZE,
};

enum HeadwayTabStops {
	HTS_TAB0=2,
	HTS_TAB1=174,
	HTS_TAB2=261,
	HTS_TAB3=348,
	HTS_END=559,
};

static const Widget _headway_list_widgets[] = {
	{   WWT_CLOSEBOX,   RESIZE_NONE,  COLOUR_GREY,     0,    10,     0,    13, STR_BLACK_CROSS,      STR_TOOLTIP_CLOSE_WINDOW},            // HLW_WIDGET_CLOSEBOX
	{    WWT_CAPTION,  RESIZE_RIGHT,  COLOUR_GREY,    11,   547,     0,    13, STR_HEADWAY_CAPTION,  STR_TOOLTIP_WINDOW_TITLE_DRAG_THIS},  // HLW_WIDGET_CAPTION
	{  WWT_STICKYBOX,     RESIZE_LR,  COLOUR_GREY,   548,   559,     0,    13, 0x0,                  STR_TOOLTIP_STICKY},                  // HLW_WIDGET_STICKY
	{      WWT_PANEL,  RESIZE_RIGHT,  COLOUR_GREY,     0,   467,    14,    20, 0x0,                  STR_NULL},                            // HLW_WIDGET_EMPTY_TOP_LEFT
	{      WWT_PANEL,     RESIZE_RB,  COLOUR_GREY,     0,   407,    21,   176, 0x0,                  STR_HEADWAY_ORDERLIST_TOOLTIP},       // HLW_WIDGET_ORDER_LIST
	{  WWT_SCROLLBAR,    RESIZE_LRB,  COLOUR_GREY,   408,   407,    14,   176, 0x0,                  STR_TOOLTIP_VSCROLL_BAR_SCROLLS_LIST},// HLW_WIDGET_FAKED_SCROLLBAR1
	{     WWT_MATRIX,    RESIZE_LRB,  COLOUR_GREY,   408,   467,    21,   176, 0x0,                  STR_HEADWAY_VEHLIST_TOOLTIP},         // HLW_WIDGET_VEH_LIST
	{  WWT_SCROLLBAR,    RESIZE_LRB,  COLOUR_GREY,   468,   467,    14,   176, 0x0,                  STR_TOOLTIP_VSCROLL_BAR_SCROLLS_LIST},// HLW_WIDGET_FAKED_SCROLLBAR2
	{     WWT_MATRIX,    RESIZE_LRB,  COLOUR_GREY,   468,   547,    14,   169, 0x0,                  STR_HEADWAY_GAPLIST_TOOLTIP},         // HLW_WIDGET_GAP_LIST
	{  WWT_SCROLLBAR,    RESIZE_LRB,  COLOUR_GREY,   548,   559,    14,   176, 0x0,                  STR_TOOLTIP_VSCROLL_BAR_SCROLLS_LIST},// HLW_WIDGET_SCROLLBAR
	{     WWT_PANEL,    RESIZE_LRTB,  COLOUR_GREY,   468,   547,   170,   176, 0x0,                  STR_NULL},                            // HLW_WIDGET_EMPTY_GAP
	{      WWT_PANEL,    RESIZE_RTB,  COLOUR_GREY,     0,   559,   177,   209, 0x0,                  STR_HEADWAY_INFO_TOOLTIP},            // HLW_WIDGET_INFO_PANEL

	{ WWT_PUSHTXTBTN,     RESIZE_TB,  COLOUR_GREY,     0,   173,   210,   221, STR_HEADWAY_INTERVAL, STR_HEADWAY_INTERVAL_TOOLTIP},        // HLW_WIDGET_INTERVAL
	{ WWT_PUSHTXTBTN,     RESIZE_TB,  COLOUR_GREY,   174,   347,   210,   221, STR_HEADWAY_AUTOMATIC,STR_HEADWAY_AUTOMATIC_TOOLTIP},       // HLW_WIDGET_AUTOMATIC
	{ WWT_PUSHTXTBTN,     RESIZE_TB,  COLOUR_GREY,   348,   523,   210,   221, STR_HEADWAY_RESET_LATENESS, STR_HEADWAY_RESET_LATENESS_TOOLTIP}, // HLW_WIDGET_RESET_LATENESS

	{ WWT_PUSHIMGBTN,     RESIZE_TB,  COLOUR_GREY,   524,   535,   210,   221, SPR_FLAG_VEH_STOPPED, STR_VEHICLE_LIST_MASS_STOP_LIST_TOOLTIP},  // HLW_WIDGET_STOP_ALL
	{ WWT_PUSHIMGBTN,     RESIZE_TB,  COLOUR_GREY,   536,   547,   210,   221, SPR_FLAG_VEH_RUNNING, STR_VEHICLE_LIST_MASS_START_LIST_TOOLTIP}, // HLW_WIDGET_START_ALL
	{      WWT_PANEL,    RESIZE_RTB,  COLOUR_GREY,   548,   547,   210,   221, 0x0,                  STR_NULL},                            // HLW_WIDGET_EMPTY_BOTTOM_RIGHT
	{  WWT_RESIZEBOX,   RESIZE_LRTB,  COLOUR_GREY,   548,   559,   210,   221, 0x0,                  STR_TOOLTIP_RESIZE},                  // HLW_WIDGET_RESIZE
	{   WIDGETS_END},
};


static void ChangeTimetableHeadwayIntervalCallback(Window *w, uint32 interval);

/**
 * Window for the headway vehicle listing.
 *
 * bitmask for w->window_number
 * 0-15 OrderID
 */
struct HeadwayListWindow : TimetableWindowBase {
	enum TextEvents { EV_CHANGE_TIMETABLE, EV_CHANGE_INTERVAL };

	enum SelectionType {
		SEL_NONE,
		SEL_ORDER,
		SEL_VEHICLE,
		SEL_GAP,
	};

	struct VehicleListEntry {
		bool empty;
		union {
			const Vehicle *veh;
			unsigned order_sub_index;
		} u;

		inline unsigned GetCurrentOrderSubIndex() const
		{
			return empty ? u.order_sub_index : u.veh->GetCurrentOrderSubIndex();
		}
	};

	typedef GUIList<VehicleListEntry> VehicleList;

	static const StringID _veh_string[];

	static VehicleList::SortFunction *const headway_sorter_funcs[];

	OrderList *order_list;
	VehicleList vehicles;
	Listing sorting;
	SelectionType sel_type;
	int sel_index;
	TextEvents text_event;

	HeadwayListWindow(const WindowDesc *desc, WindowNumber window_number)
		: TimetableWindowBase(desc, window_number), order_list(OrderList::GetIfValid(window_number)), sel_type(SEL_NONE), sel_index(-1)
	{
		assert(this->order_list != NULL);
		assert(this->order_list->IsShared());

		this->owner = this->order_list->GetFirstSharedVehicle()->owner;

		this->resize.step_width = 1;
		this->vscroll.SetCapacity(12);
		this->resize.step_height = PLY_WND_PRC__SIZE_OF_ROW_TINY;

		this->widget[HLW_WIDGET_VEH_LIST].data = (this->vscroll.GetCapacity() << MAT_ROW_START) + (1 << MAT_COL_START);
		this->widget[HLW_WIDGET_GAP_LIST].data = (this->vscroll.GetCapacity() << MAT_ROW_START) + (1 << MAT_COL_START);

		this->HideWidget(HLW_WIDGET_FAKED_SCROLLBAR1);
		this->HideWidget(HLW_WIDGET_FAKED_SCROLLBAR2);

		sorting.order = false;
		sorting.criteria = 0;

		this->vehicles.SetSortFuncs(this->headway_sorter_funcs);
		this->vehicles.SetListing(sorting);
		this->vehicles.ForceRebuild();
		this->vehicles.NeedResort();

		this->FindWindowPlacementAndResize(desc);
	}

	void BuildVehicleList()
	{
		int empty[INVALID_VEH_ORDER_ID*2];

		if (!this->vehicles.NeedRebuild()) return;

		if (this->order_list->IsCompleteTimetable()) {
			Ticks interval = max(1U, this->order_list->GetTimetableDurationIncomplete() / this->order_list->GetNumRunningVehicles());

			for (int i = 0; i < order_list->GetNumOrders() * 2; i += 2) {
				const Order *order = this->order_list->GetOrderAt(i / 2);
				empty[i] = order->wait_time / interval + 1;

				order = order->next;
				if (order == NULL) order = this->order_list->GetFirstOrder();
				empty[i+1] = order->travel_time / interval + 1;
			}
		} else {
			for (int i = 0; i < order_list->GetNumOrders() * 2; ++i) empty[i] = 1;
		}

		const Vehicle *sel_veh = NULL;
		if (this->sel_type == SEL_VEHICLE || this->sel_type == SEL_GAP) sel_veh = this->vehicles[this->sel_index].u.veh;

		this->vehicles.Clear();

		/* Add all vehicles from this vehicle's shared order list */
		for (const Vehicle *v = this->order_list->GetFirstSharedVehicle(); v != NULL; v = v->NextShared()) {
			if (v->IsStoppedInDepot()) continue;

			VehicleListEntry *entry = this->vehicles.Append();
			entry->empty = false;
			entry->u.veh = v;
			--empty[v->GetCurrentOrderSubIndex()];
			if (v == sel_veh) this->sel_index = this->vehicles.Length() - 1;
		}

		for (int i = 0; i < order_list->GetNumOrders() * 2; ++i) {
			while (empty[i] > 0) {
				VehicleListEntry *entry = this->vehicles.Append();
				entry->empty = true;
				entry->u.order_sub_index = i;
				--empty[i];
			}
		}

		this->vehicles.Compact();
		this->vehicles.RebuildDone();
	}

	void SortVehicleList()
	{
		const Vehicle *v = (this->sel_type == SEL_VEHICLE || this->sel_type == SEL_GAP) ? this->vehicles[this->sel_index].u.veh : NULL;

		if (this->vehicles.Sort() && v != NULL) {
			for (unsigned i = 0; i < this->vehicles.Length(); ++i) {
				if (!this->vehicles[i].empty && this->vehicles[i].u.veh == v) {
					this->sel_index = i;
					break;
				}
			}
		}
	}

	void DrawLatenessStatistic(int left, int right, int y, int32 min_lateness, int max_lateness)
	{
		StringID str;

		SetTimetableParams(0, 1, abs(min_lateness));
		SetTimetableParams(2, 3, abs(max_lateness));

		if (min_lateness < 0 && max_lateness < 0) str = STR_HEADWAY_ALL_EARLY;
		else if (min_lateness < 0) str = STR_HEADWAY_EARLY_AND_LATE;
		else if (min_lateness == 0 && max_lateness == 0) str = STR_HEADWAY_ALL_ON_TIME;
		else str = STR_HEADWAY_ALL_LATE;

		DrawString(left, right, y, str, TC_FROMSTRING);
	}

	void DrawAllInfoPanel(int y)
	{
		SetTimetableParams(0, 1, this->order_list->GetTimetableDurationIncomplete());
		DrawString(HTS_TAB0, HTS_TAB1, y, this->order_list->IsCompleteTimetable() ? STR_HEADWAY_TOTAL_TIME : STR_HEADWAY_TOTAL_TIME_INCOMPLETE, TC_BLACK);

		int32 min_lateness, max_lateness;
		this->order_list->CalculateLatenessStatistics(&min_lateness, &max_lateness);
		DrawLatenessStatistic(HTS_TAB1, HTS_END, y, min_lateness, max_lateness);
	}

	void DrawVehicleInfoPanel(int y, const Vehicle *v)
	{
		bool travelling = (!v->current_order.IsType(OT_LOADING) || v->current_order.GetNonStopType() == ONSF_STOP_EVERYWHERE);

		SetDParam(0, this->_veh_string[v->type]);
		SetDParam(1, v->unitnumber);
		Ticks timetabled = travelling ? v->current_order.travel_time : v->current_order.wait_time;
		Ticks scheduled = DateToTicks() + timetabled - v->current_order_time - v->lateness_counter;

		/* Scheduled arrival/departure */
		SetArrivalDepartParams(2, 3, scheduled);
		/* Expected arrival/departure */
		if (travelling || v->lateness_counter > 0) {
			SetArrivalDepartParams(4, 5, DateToTicks() + max(0, timetabled - (int)v->current_order_time));
		} else {
			SetArrivalDepartParams(4, 5, scheduled);
		}

		DrawString(HTS_TAB0, HTS_TAB2, y, travelling ? STR_HEADWAY_VEH_ARRIVAL : STR_HEADWAY_VEH_DEPARTURE, TC_FROMSTRING);

		if (!HasBit(v->vehicle_flags, VF_TIMETABLE_STARTED)) {
			DrawString(HTS_TAB2, HTS_END, y, STR_TIMETABLE_STATUS_NOT_STARTED, TC_BLACK);
		} else if (v->lateness_counter == 0 || (!_settings_client.gui.timetable_unit && v->lateness_counter / DAY_TICKS == 0)) {
			DrawString(HTS_TAB2, HTS_END, y, STR_TIMETABLE_STATUS_ON_TIME, TC_BLACK);
		} else {
			SetTimetableParams(0, 1, abs(v->lateness_counter));
			DrawString(HTS_TAB2, HTS_END, y, v->lateness_counter < 0 ? STR_TIMETABLE_STATUS_EARLY : STR_TIMETABLE_STATUS_LATE, TC_BLACK);
		}
	}

	void DrawOrderInfoPanel(int y)
	{
		assert(this->sel_type == SEL_ORDER);
		assert(this->sel_index >= 0);

		unsigned i;
		for (i = 0; i < this->vehicles.Length(); ++i) {
			if (this->vehicles[i].GetCurrentOrderSubIndex() == (unsigned)sel_index) break;
		}
		assert(i < this->vehicles.Length());

		if (this->vehicles[i].empty) {
			DrawString(HTS_TAB0, HTS_END, y, STR_HEADWAY_ORDER_NO_VEHICLE, TC_BLACK);
		} else if (i == this->vehicles.Length() - 1 || this->vehicles[i+1].empty || this->vehicles[i+1].GetCurrentOrderSubIndex() > (unsigned)sel_index) {
			DrawVehicleInfoPanel(y, this->vehicles[i].u.veh);
		} else {
			int32 min_lateness = INT32_MAX;
			int32 max_lateness = INT32_MIN;
			unsigned num_vehicles = 0;

			while (i < this->vehicles.Length() && !this->vehicles[i].empty && this->vehicles[i].GetCurrentOrderSubIndex() == (unsigned)sel_index) {
				min_lateness = min(min_lateness, this->vehicles[i].u.veh->lateness_counter);
				max_lateness = max(max_lateness, this->vehicles[i].u.veh->lateness_counter);
				++num_vehicles;
				++i;
			}
			SetDParam(0, num_vehicles);
			DrawString(HTS_TAB0, HTS_TAB1, y, STR_HEADWAY_ORDER_NUM_VEHICLES, TC_BLACK);
			DrawLatenessStatistic(HTS_TAB1, HTS_END, y, min_lateness, max_lateness);
		}
	}

	const Vehicle *GetVehicleBeforeGap(int index)
	{
		unsigned num_entries = this->vehicles.Length();
		int j = (index + num_entries - 1) % num_entries;
		while (this->vehicles[j].empty) j = (j + num_entries - 1) % num_entries;
		return this->vehicles[j].u.veh;
	}

	void DrawGapInfoPanel(int y)
	{
		assert(this->sel_type == SEL_GAP);
		assert(this->sel_index >= 0);

		const Vehicle *v1 = GetVehicleBeforeGap(this->sel_index);
		const Vehicle *v2 = this->vehicles[this->sel_index].u.veh;

		Ticks actual_gap = v2->GetTimetablePosition(true, false) - v1->GetTimetablePosition(true, false);
		if (actual_gap <= -this->order_list->GetTimetableTotalDuration()/2) actual_gap += this->order_list->GetTimetableTotalDuration();
		SetTimetableParams(0, 1, actual_gap);
		DrawString(2, y, STR_HEADWAY_REAL_GAP, TC_BLACK);

		SetDParam(0, this->_veh_string[v1->type]);
		SetDParam(1, v1->unitnumber);
		if (v1->lateness_counter == 0) SetDParam(2, STR_HEADWAY_VEH_ON_TIME);
		else if (v1->lateness_counter < 0) SetDParam(2, STR_HEADWAY_VEH_EARLY);
		else SetDParam(2, STR_HEADWAY_VEH_LATE);
		SetTimetableParams(3, 4, abs(v1->lateness_counter));

		SetDParam(5, this->_veh_string[v2->type]);
		SetDParam(6, v2->unitnumber);
		if (v2->lateness_counter == 0) SetDParam(7, STR_HEADWAY_VEH_ON_TIME);
		else if (v2->lateness_counter < 0) SetDParam(7, STR_HEADWAY_VEH_EARLY);
		else SetDParam(7, STR_HEADWAY_VEH_LATE);
		SetTimetableParams(8, 9, abs(v2->lateness_counter));

		DrawString(HTS_TAB1, HTS_END, y, STR_HEADWAY_VEHICLE_PAIR, TC_FROMSTRING);
	}

	void DrawInfoPanel()
	{
		StringID str;
		int y = widget[HLW_WIDGET_INFO_PANEL].top + 2;

		if (order_list->GetHeadwayFlags() == 0) str = STR_HEADWAY_INACTIVE;
		else if (HasBit(order_list->GetHeadwayFlags(), HWF_AUTOFILL)) str = STR_HEADWAY_AUTOFILL_AND_INITIALIZE;
		else if (order_list->IsHeadwayActive() && HasBit(order_list->GetHeadwayFlags(), HWF_INITIALIZE)) str = STR_HEADWAY_REINITIALIZING;
		else if (HasBit(order_list->GetHeadwayFlags(), HWF_INITIALIZE)) str = STR_HEADWAY_INITIALIZING;
		else if (HasBit(order_list->GetHeadwayFlags(), HWF_FIRST_ROUND)) str = STR_HEADWAY_FIRST_ROUND;
		else if (HasBit(order_list->GetHeadwayFlags(), HWF_AUTOMATIC)) str = STR_HEADWAY_IS_AUTOMATIC;
		else str = STR_HEADWAY_ACTIVE;

		DrawString(HTS_TAB0, HTS_TAB1, y, str, TC_FROMSTRING);

		if (order_list->GetHeadwayFlags() != 0) {
			if (!HasBit(order_list->GetHeadwayFlags(), HWF_AUTOFILL)) {
				SetTimetableParams(0, 1, order_list->GetHeadwayInterval());
				DrawString(HTS_TAB1, HTS_TAB3, y, STR_HEADWAY_SHOW_INTERVAL, TC_FROMSTRING);
			} else {
				DrawString(HTS_TAB1, HTS_TAB3, y, STR_HEADWAY_UNKNOWN_INTERVAL, TC_FROMSTRING);
			}

			if (this->order_list->GetHeadwayLeader() == NULL) {
				SetDParam(0, STR_HEADWAY_NO_LEADER);
			} else {
				SetDParam(0, this->_veh_string[this->order_list->GetHeadwayLeader()->type]);
				SetDParam(1, this->order_list->GetHeadwayLeader()->unitnumber);
			}

			DrawString(HTS_TAB3, HTS_END, y, STR_HEADWAY_SHOW_LEADER, TC_FROMSTRING);
		}

		y += 10;

		if (this->sel_type == SEL_VEHICLE && !this->vehicles[this->sel_index].empty) {
			DrawVehicleInfoPanel(y, this->vehicles[this->sel_index].u.veh);
		} else if (this->sel_type == SEL_ORDER) {
			DrawOrderInfoPanel(y);
		} else if (this->sel_type == SEL_GAP) {
			DrawGapInfoPanel(y);
		} else {
			DrawAllInfoPanel(y);
		}
	}

	void DrawOrderEntry(int y, unsigned i, bool draw_top_line)
	{
		Widget &wi = this->widget[HLW_WIDGET_ORDER_LIST];

		if (draw_top_line) {
			GfxFillRect(wi.left + 1, y - 1, wi.right - 1, y - 1, _colour_gradient[wi.colour & 0xF][4]);
		}

		int index = this->vehicles[i].GetCurrentOrderSubIndex();

		if (index % 2 == 0) {
			const Order *order = this->order_list->GetOrderAt(index / 2);
			DrawOrderString(this->order_list->GetFirstSharedVehicle(), order, index / 2, y + 2, (this->sel_type == SEL_ORDER && this->sel_index == index), true, wi.left + 2, wi.right - 4, false);
		} else {
			const Order *order = this->order_list->GetOrderAt(((index + 1) / 2) % this->order_list->GetNumOrders());
			StringID string;

			if (order->IsType(OT_CONDITIONAL)) {
				string = STR_TIMETABLE_NO_TRAVEL;
			} else if (order->travel_time == 0) {
				string = STR_TIMETABLE_TRAVEL_NOT_TIMETABLED;
			} else {
				SetTimetableParams(0, 1, order->travel_time);
				string = STR_TIMETABLE_TRAVEL_FOR;
			}

			DrawString(wi.left + 2, wi.right - 4, y + 2, string, (this->sel_type == SEL_ORDER && this->sel_index == index) ? TC_WHITE : TC_BLACK);
		}

		GfxFillRect(wi.left + 1, y, wi.right - 1, y, _colour_gradient[wi.colour & 0xF][6]);
	}

	void DrawVehicleEntry(int y, int i, const Vehicle *v)
	{
		TextColour colours[][3] = {
			{ TC_BLACK, TextColour(_colour_gradient[COLOUR_GREEN][3] | IS_PALETTE_COLOUR), TextColour(_colour_gradient[COLOUR_RED][3] | IS_PALETTE_COLOUR) },
			{ TC_WHITE, TextColour(_colour_gradient[COLOUR_GREEN][7] | IS_PALETTE_COLOUR), TextColour(_colour_gradient[COLOUR_RED][7] | IS_PALETTE_COLOUR) }
		};

		bool selected = (this->sel_type == SEL_VEHICLE && this->sel_index == (int)i) ||
		                (this->sel_type == SEL_ORDER && (int)v->GetCurrentOrderSubIndex() == this->sel_index) ||
		                (this->sel_type == SEL_GAP &&
		                  (this->sel_index == i ||
		                   this->sel_index == (i + 1) % (int)this->vehicles.Length()));

		if (!selected && this->sel_type == SEL_GAP) {
			int j = (i + 1) % this->vehicles.Length();
			while (this->vehicles[j].empty) j = (j + 1) % this->vehicles.Length();
			if (this->sel_index == j) selected = true;
		}

		int colour_index = 0;
		if (v->lateness_counter >= DAY_TICKS) colour_index = 2;
		else if (v->lateness_counter <= -DAY_TICKS) colour_index = 1;

		SetDParam(0, this->_veh_string[v->type]);
		SetDParam(1, v->unitnumber);
		SetDParam(2, (v == this->order_list->GetHeadwayLeader()) ? STR_CARET : STR_EMPTY);
		SetDParam(3, HasBit(v->vehicle_flags, VF_AUTOFILL_TIMETABLE) ? STR_AT : STR_EMPTY);

		DrawString(this->widget[HLW_WIDGET_VEH_LIST].left + 5, this->widget[HLW_WIDGET_VEH_LIST].right - 2, y + 2, STR_HEADWAY_VEH_ENTRY, colours[(int)selected][colour_index]);
	}

	void DrawGapEntry(int y, int i, const Vehicle *v, const Vehicle **prev_v)
	{
		if (!this->order_list->IsCompleteTimetable()) return;

		Ticks gap = v->GetTimetablePosition(true, true) - (*prev_v)->GetTimetablePosition(true, true);
		if (gap <= -this->order_list->GetTimetableTotalDuration()/2) gap += this->order_list->GetTimetableTotalDuration();
		SetTimetableParams(0, 1, gap);
		// (_settings_client.gui.timetable_unit == TTU_HHMMSS || _settings_client.gui.timetable_unit == TTU_HHMM) ? abs(gap) :
		TextColour colour = (gap >= 0) ? TC_BLACK : TextColour(_colour_gradient[COLOUR_RED][3] | IS_PALETTE_COLOUR);
		if (this->sel_type == SEL_GAP && this->sel_index == (int)i) colour = (gap >= 0) ? TC_WHITE : TextColour(_colour_gradient[COLOUR_RED][7] | IS_PALETTE_COLOUR);
		DrawString(this->widget[HLW_WIDGET_GAP_LIST].left + 5, this->widget[HLW_WIDGET_GAP_LIST].right - 2, y + 2, STR_HEADWAY_GAP, colour);

		*prev_v = v;
	}

	virtual void OnPaint()
	{
		assert(this->order_list != NULL);
		assert(this->order_list->IsShared());

		this->BuildVehicleList();
		this->SortVehicleList();

		this->vscroll.SetCount(this->vehicles.Length());

		assert(this->vehicles.Length() >= 2);

		bool complete = this->order_list->IsCompleteTimetable();

		this->SetWidgetDisabledState(HLW_WIDGET_INTERVAL, !complete && this->sel_type != SEL_VEHICLE);
		this->SetWidgetDisabledState(HLW_WIDGET_AUTOMATIC, (order_list->GetHeadwayFlags() & (HWFB_ACTIVE | HWFB_INITIALIZE)) == 0);
		this->SetWidgetDisabledState(HLW_WIDGET_RESET_LATENESS, (order_list->GetHeadwayFlags() & (HWFB_ACTIVE | HWFB_INITIALIZE)) == 0 || HasBit(order_list->GetHeadwayFlags(), HWF_FIRST_ROUND));

		this->SetWidgetLoweredState(HLW_WIDGET_AUTOMATIC, HasBit(this->order_list->GetHeadwayFlags(), HWF_AUTOMATIC));

		SetDParam(0, this->order_list->GetNumVehicles());
		this->DrawWidgets();

		if (!complete && this->sel_type == SEL_GAP) {
			this->sel_type = SEL_NONE;
			this->sel_index = -1;
		}

		const int offset =  this->widget[HLW_WIDGET_ORDER_LIST].top - this->widget[HLW_WIDGET_GAP_LIST].top;
		int y = this->widget[HLW_WIDGET_GAP_LIST].top;

		unsigned i;
		for (i = (this->vscroll.GetPosition() + this->vehicles.Length() - 1) % this->vehicles.Length(); this->vehicles[i].empty; i = (i + this->vehicles.Length() - 1) % this->vehicles.Length()) {}
		const Vehicle *prev_v =	this->vehicles[i].u.veh;

		for (i = this->vscroll.GetPosition(); i < this->vehicles.Length() && y < widget[HLW_WIDGET_GAP_LIST].bottom; ++i) {
			if (y == this->widget[HLW_WIDGET_GAP_LIST].top || this->vehicles[i-1].GetCurrentOrderSubIndex() != this->vehicles[i].GetCurrentOrderSubIndex()) {
				DrawOrderEntry(y + offset, i, (y != this->widget[HLW_WIDGET_GAP_LIST].top));
			}

			if (!this->vehicles[i].empty) {
				const Vehicle *v = this->vehicles[i].u.veh;
				DrawVehicleEntry(y + offset, i, v);
				DrawGapEntry(y, i, v, &prev_v);
			}

			y += resize.step_height;
		}

		DrawInfoPanel();
	}

	virtual void OnClick(Point pt, int widget)
	{
		switch (widget) {
			case HLW_WIDGET_GAP_LIST: { // Matrix to show gaps
				if (!this->order_list->IsCompleteTimetable()) break;
				/* Fall-through */
			case HLW_WIDGET_VEH_LIST:   // Matrix to show vehicles
			case HLW_WIDGET_ORDER_LIST: // Matrix to show order
				SelectionType prev_sel_type = this->sel_type;
				int prev_sel_index = this->sel_index;
				int id_v = (pt.y - this->widget[widget].top) / this->resize.step_height;

				if (!_ctrl_pressed) {
					this->sel_type = SEL_NONE;
					this->sel_index = -1;
				}

				if (id_v >= this->vscroll.GetCapacity()) return; // click out of bounds

				id_v += this->vscroll.GetPosition();

				if ((unsigned)id_v >= this->vehicles.Length()) { // click out of list bound
					if (widget != HLW_WIDGET_ORDER_LIST) return;
					id_v = this->vehicles.Length() - 1;
				}

				if (this->vehicles[id_v].empty && !(widget == HLW_WIDGET_ORDER_LIST)) return; // no vehicle in this cell

				if (_ctrl_pressed) {
					switch (widget) {
						case HLW_WIDGET_ORDER_LIST: {
							int selected = this->vehicles[id_v].GetCurrentOrderSubIndex();
							VehicleOrderID real = ((selected + 1) / 2) % order_list->GetNumOrders();
							const Order *order = order_list->GetOrderAt(real);
							if (selected % 2 == 1) {
								if (order->IsType(OT_CONDITIONAL)) break;
							} else {
								if ((!order->IsType(OT_GOTO_STATION) || (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION)) && !order->IsType(OT_CONDITIONAL)) break;
							}
							text_event = EV_CHANGE_TIMETABLE;
							ShowChangeTimetableWindow(order, selected, order_list->GetFirstSharedVehicle(), (selected % 2 == 1), this);
						}	break;

						case HLW_WIDGET_VEH_LIST: {
							const Vehicle *v = this->vehicles[id_v].u.veh;
							ShowVehicleViewWindow(v);
						} break;

						case HLW_WIDGET_GAP_LIST: {
							const Vehicle *v1 = GetVehicleBeforeGap(id_v);
							const Vehicle *v2 = this->vehicles[id_v].u.veh;
							ShowVehicleViewWindow(v1);
							ShowVehicleViewWindow(v2);
						} break;
					}
				} else {
					switch (widget) {
						case HLW_WIDGET_ORDER_LIST:
							this->sel_type = SEL_ORDER;
							this->sel_index = this->vehicles[id_v].GetCurrentOrderSubIndex();
							break;

						case HLW_WIDGET_VEH_LIST:
							this->sel_type = SEL_VEHICLE;
							this->sel_index = id_v;
							break;

						case HLW_WIDGET_GAP_LIST:
							this->sel_type = SEL_GAP;
							this->sel_index = id_v;
							break;
					}
				}

				if (!_ctrl_pressed && this->sel_type == prev_sel_type && this->sel_index == prev_sel_index) {
					this->sel_type = SEL_NONE;
					this->sel_index = -1;
				}
			} break;


			case HLW_WIDGET_INTERVAL: {
				Ticks interval = order_list->GetHeadwayInterval();
				bool complete = this->order_list->IsCompleteTimetable();

				if (!complete) {
					assert(this->sel_type == SEL_VEHICLE);
					assert(this->sel_index >= 0);
					assert((unsigned)this->sel_index < this->vehicles.Length());

					DoAutofillCommand(this->vehicles[this->sel_index].u.veh, true, true, 0);
					ChangeTimetableHeadwayIntervalCallback(this, 0xffffffff);
				} else {
					if (interval == INVALID_TICKS || _ctrl_pressed) {
						Ticks duration = order_list->GetTimetableTotalDuration();
						if (duration > 0) interval = duration / order_list->GetNumRunningVehicles();
					}

					if (_settings_client.gui.timetable_unit == TTU_DAYS || _settings_client.gui.timetable_unit == TTU_TICKS) {
						StringID caption = STR_EMPTY;
						StringID current = STR_EMPTY;

						switch (_settings_client.gui.timetable_unit) {
							case TTU_DAYS:
								caption = STR_HEADWAY_INTERVAL_DAYS;
								interval /= DAY_TICKS;
								break;
							case TTU_TICKS:
								caption = STR_HEADWAY_INTERVAL_TICKS;
								break;
							default: NOT_REACHED();
						}

						if (interval != 0) {
							SetDParam(0, interval);
							current = STR_JUST_INT;
						}

						text_event = EV_CHANGE_INTERVAL;
						ShowQueryString(current, caption, 31, 150, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
					} else { // HHMM or HHMMSS
						interval = TicksToTime(interval);
						ShowSetDateHHMMSSWindow(this, this->window_number | (1 << 31), interval, (_settings_client.gui.timetable_unit == TTU_HHMMSS), ChangeTimetableHeadwayIntervalCallback);
					}
				}
			} break;

			case HLW_WIDGET_AUTOMATIC:
				DoCommandP(0, this->order_list->index | (HasBit(this->order_list->GetHeadwayFlags(), HWF_AUTOMATIC) ? 0 : (1 << 31)), 0, CMD_SET_HEADWAY_AUTOMATIC);
				break;

			case HLW_WIDGET_RESET_LATENESS:
				DoCommandP(0, this->order_list->index | (_ctrl_pressed ? (1 << 17) : (1 << 16)) | (1 << 18), 0, CMD_SET_HEADWAY_LATENESS);
				break;

			case HLW_WIDGET_STOP_ALL:
			case HLW_WIDGET_START_ALL:
				DoCommandP(0, this->order_list->GetFirstSharedVehicle()->index, VLW_SHARED_ORDERS | (1 << 6) | (widget == HLW_WIDGET_START_ALL ? (1 << 5) : 0) | this->order_list->GetFirstSharedVehicle()->type, CMD_MASS_START_STOP);
				break;
		}
		this->SetDirty(); // TODO
	}

	virtual void OnTick()
	{
		if (_pause_mode != PM_UNPAUSED) return;
		if (this->vehicles.NeedResort()) {
			CompanyID owner = this->owner;

			DEBUG(misc, 3, "Periodic resort headway list player %d, order list ID %d", owner, this->window_number);
			this->SetDirty();
		}
	}

	virtual void OnQueryTextFinished(char *str)
	{
		if (str == NULL) return;

		switch (this->text_event) {
			case EV_CHANGE_TIMETABLE: {
				ChangeTimetableCallback(this->change_tt_veh, this->change_tt_order, str);
				this->change_tt_veh = INVALID_VEHICLE;
				this->change_tt_order = -1;
			} break;

			case EV_CHANGE_INTERVAL: {
				uint64 interval = StrEmpty(str) ? 0 : strtoul(str, NULL, 10);
				if (_settings_client.gui.timetable_unit == TTU_DAYS) interval *= DAY_TICKS;
				if (interval > (1U << 20) - 1) return;

				ChangeTimetableHeadwayIntervalCallback(this, interval);
			} break;

			default:
				NOT_REACHED();
		}
	}

	virtual void OnResize(Point new_size, Point delta)
	{
		this->vscroll.SetCapacity(this->vscroll.GetCapacity() + delta.y / (int)this->resize.step_height);
		this->widget[HLW_WIDGET_VEH_LIST].data = (this->vscroll.GetCapacity() << MAT_ROW_START) + (1 << MAT_COL_START);
		this->widget[HLW_WIDGET_GAP_LIST].data = (this->vscroll.GetCapacity() << MAT_ROW_START) + (1 << MAT_COL_START);
	}

	virtual void OnInvalidateData(int data)
	{
		assert(order_list->GetNumRunningVehicles() >= 2);

		this->vehicles.ForceRebuild();
	}
};

static void ChangeTimetableHeadwayIntervalCallback(Window *w, uint32 interval)
{
	uint32 p1 = w->window_number;
	HeadwayListWindow *hlw = dynamic_cast<HeadwayListWindow*>(w);
	assert(hlw);

	if (hlw->sel_type == HeadwayListWindow::SEL_VEHICLE && hlw->sel_index >= 0 && (unsigned)hlw->sel_index < hlw->vehicles.Length()) {
		SB(p1, 16, 16, hlw->vehicles[hlw->sel_index].u.veh->index);
	} else {
		SB(p1, 16, 16, INVALID_VEHICLE);
	}

	if (interval != 0xffffffff && (_settings_client.gui.timetable_unit == TTU_HHMM || _settings_client.gui.timetable_unit == TTU_HHMMSS)) interval = TimeToTicks(interval);
	uint32 p2 = minu(interval, ((1U << 20) - 1));

	DoCommandP(0, p1, p2, CMD_SET_HEADWAY_INTERVAL | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE));

	hlw->SetDirty(); // TODO
}

static WindowDesc _headway_list_desc(
	WDP_AUTO, WDP_AUTO, 560, 222, 560, 222,
	WC_HEADWAY_LIST, WC_VEHICLE_TIMETABLE,
	WDF_STD_TOOLTIPS | WDF_STD_BTN | WDF_DEF_WIDGET | WDF_UNCLICK_BUTTONS | WDF_STICKY_BUTTON | WDF_RESIZABLE,
	_headway_list_widgets
);

/** Sort by position in the headway (primary: order index, secondary current order time) */
static int CDECL VehicleHeadwayPositionSorter(const HeadwayListWindow::VehicleListEntry *a, const HeadwayListWindow::VehicleListEntry *b)
{
	int r = a->GetCurrentOrderSubIndex() - b->GetCurrentOrderSubIndex();

	if (r == 0) {
		if (a->empty && b->empty) return 0;
		if (a->empty) return 1;
		if (b->empty) return -1;
		return a->u.veh->current_order_time - b->u.veh->current_order_time;
	}

	return r;
}

const StringID HeadwayListWindow::_veh_string[] = {
	STR_TRAIN_ICON_AND_NUMBER,
	STR_LORRY_ICON_AND_NUMBER,
	STR_SHIP_ICON_AND_NUMBER,
	STR_PLANE_ICON_AND_NUMBER
};

HeadwayListWindow::VehicleList::SortFunction *const HeadwayListWindow::headway_sorter_funcs[] = {
	VehicleHeadwayPositionSorter
};

void ShowHeadwayListWindow(uint16 order_list)
{
	if (!OrderList::IsValidID(order_list)) return;

	AllocateWindowDescFront<HeadwayListWindow>(&_headway_list_desc, order_list);
}
